﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading.Tasks;
using System.Timers;

using System.IO;
using System.Threading; //CancelationToken
using Google.Apis;
using Google.Apis.Download;
using Google.Apis.Auth.OAuth2;
using Google.Apis.YouTubeReporting.v1; //Reporting API Auth
using Google.Apis.YouTubeReporting.v1.Data; //Reporting API Data

using Google.Apis.Services; //Base Client Service
using Google.Apis.Util.Store; //FileDataStore needs this

using Google.Apis.YouTube.v3; //Data API Auth
using Google.Apis.YouTube.v3.Data; //Data API Data

using System.Data.SqlClient; //Db Store

namespace YouTubeReportingConsole
{
    class Program
    {
        static void Main(string[] args)
        {
            //App supports scheduling and downloading YouTube channel reports including branded accounts.  
            //To support content owner reports collect id and specify OnBehalfOfContentOwner
            //Console.ReadLine();

            if (args.Length > 0)
            {
                switch (args[0].ToLower())
                {
                    case "reporttypeslist": //ReportTypesList returns a list of report types that are available
                        ReportTypesResource.ListRequest reportTypesList = ytReportingService.ReportTypes.List();
                        ListReportTypesResponse reportTypesListResponse = reportTypesList.Execute();
                        foreach (ReportType reportType in reportTypesListResponse.ReportTypes)
                        {
                            WriteReportType(reportType, "Report Type List");
                        }
                        break;
                    case "jobcreate": //JobCreate creates jobs of the specified types.  If the second argument equals "All", then a job is created for all availale report types.
                        for (int i = 1; i < args.Length; i++)
                        {
                            if (args[i].ToLower() == "all")
                            { //Create a job for all available report types
                                ListReportTypesResponse reportTypesAvailableResponse = ytReportingService.ReportTypes.List().Execute();
                                foreach (ReportType reportType in reportTypesAvailableResponse.ReportTypes)
                                {
                                    Job createdjob = CreateJob(reportType.Id);
                                    WriteJob(createdjob, "Job Created");
                                }
                                break;
                            }
                            else
                            { //Create job with id equal to the argument value
                                Job createdjob = CreateJob(args[i].ToLower());
                                WriteJob(createdjob, "Job Created");
                            }  
                        }
                        break;
                    case "jobdelete": //Delete job.  If the second argument is "All" then all jobs are deleted. If second argument is a job id, then specified job is deleted.
                        for (int i = 1; i < args.Length; i++)
                        {
                            if (args[i].ToLower() == "all")
                            { //Delete all jobs
                                ListJobsResponse jobDeleteList = ytReportingService.Jobs.List().Execute();
                                foreach (Job Job in jobDeleteList.Jobs)
                                {
                                    ytReportingService.Jobs.Delete(Job.Id).Execute();
                                    Console.WriteLine("Job Deleted: " + Job.Name);
                                }
                                break;
                            }
                            else
                            { //Delete each job with id equal to the argument value
                                Job Job = ytReportingService.Jobs.Get(args[i]).Execute();
                                ytReportingService.Jobs.Delete(args[i]).Execute();
                                Console.WriteLine("Job Deleted: " + Job.Name);
                            }
                        }
                        break;
                    case "joblist": //Return a list of all jobs that have been created.
                        ListJobsResponse jobListResponse = ytReportingService.Jobs.List().Execute();
                        if (jobListResponse.Jobs == null)
                        {
                            Console.WriteLine("No Jobs Returned");
                        }
                        else
                        {
                            foreach (Job Job in jobListResponse.Jobs)
                            {
                                WriteJob(Job, "Job List");
                            }
                        }
                        break;
                    case "videolist": //Returns a list of videos and stores published date, title, and description in generic object table.
                        //VideoList "Data Source=RG00-SQL01;Integrated Security=true; Initial Catalog=Control; MultipleActiveResultSets=True;"
                        SearchResource.ListRequest searchList = ytService.Search.List("snippet");
                        searchList.Type = "video";
                        searchList.ForMine = true;
                        searchList.MaxResults = 50; //Pagination required
                        string PageToken = "FirstRequest";
                        while (String.IsNullOrEmpty(PageToken) == false) //Max of 50 items returns per request.  Set PageToken to NextPageToken to get next rows.
                        {
                            if (PageToken != "FirstRequest")
                            {
                                searchList.PageToken = PageToken;
                            }
                            SearchListResponse searchListResponse = searchList.Execute();
                            if (searchListResponse.Items == null)
                            {
                                Console.WriteLine("No Videos Returned");
                            }
                            else
                            {
                                SqlConnection conn = OpenConnection(args[1]);
                                VerifySchemaYouTubeObject(conn);
                                foreach (var Item in searchListResponse.Items)
                                {
                                    RecordObject(conn, "video", Item.Id.VideoId, (DateTime)Item.Snippet.PublishedAt, Item.Snippet.Title, Item.Snippet.Description);
                                }
                            }
                            PageToken = searchListResponse.NextPageToken;
                        }
                        break;
                    case "channellist": //Returns a list of channels and stores published date, title, and description in generic object table.
                        //ChannelList "Data Source=RG00-SQL01;Integrated Security=true; Initial Catalog=Control; MultipleActiveResultSets=True;"
                        ChannelsResource.ListRequest channelList = ytService.Channels.List("snippet");
                        channelList.Mine = true;
                        channelList.MaxResults = 50; //Pagination required
                        string ChannelPageToken = "FirstRequest";
                        while (String.IsNullOrEmpty(ChannelPageToken) == false) //Max of 50 items returns per request.  Set PageToken to NextPageToken to get next rows.
                        {
                            if (ChannelPageToken != "FirstRequest")
                            {
                                channelList.PageToken = ChannelPageToken;
                            }
                            ChannelListResponse channelListResponse = channelList.Execute();
                            if (channelListResponse.Items == null)
                            {
                                Console.WriteLine("No Channels Returned");
                            }
                            else
                            {
                                SqlConnection conn = OpenConnection(args[1]);
                                VerifySchemaYouTubeObject(conn);
                                foreach (var Item in channelListResponse.Items)
                                {
                                    RecordObject(conn, "channel", Item.Id, (DateTime)Item.Snippet.PublishedAt, Item.Snippet.Title, Item.Snippet.Description);
                                }
                            }
                            ChannelPageToken = channelListResponse.NextPageToken;
                        }
                        break;
                    case "playlistlist": //Returns a list of playlist and stores published date, title, and description in generic object table.
                        //PlaylistList "Data Source=RG00-SQL01;Integrated Security=true; Initial Catalog=Control; MultipleActiveResultSets=True;"
                        PlaylistsResource.ListRequest playlistList = ytService.Playlists.List("snippet");
                        playlistList.Mine = true;
                        playlistList.MaxResults = 50; //Pagination required
                        string PlaylistPageToken = "FirstRequest";
                        while (String.IsNullOrEmpty(PlaylistPageToken) == false) //Max of 50 items returns per request.  Set PageToken to NextPageToken to get next rows.
                        {
                            if (PlaylistPageToken != "FirstRequest")
                            {
                                playlistList.PageToken = PlaylistPageToken;
                            }
                            PlaylistListResponse playlistListResponse = playlistList.Execute();
                            if (playlistListResponse.Items == null)
                            {
                                Console.WriteLine("No Playlists Returned");
                            }
                            else
                            {
                                SqlConnection conn = OpenConnection(args[1]);
                                VerifySchemaYouTubeObject(conn);
                                foreach (var Item in playlistListResponse.Items)
                                {
                                    RecordObject(conn, "playlist", Item.Id, (DateTime)Item.Snippet.PublishedAt, Item.Snippet.Title, Item.Snippet.Description);
                                }
                            }
                            PlaylistPageToken = playlistListResponse.NextPageToken;
                        }
                        break;
                    case "reportlist": //Collect available reports.  If second argument is "All" then reports for all jobs are collected.  If second argument is a job id, then reports for specified job id are collected.
                                       //Reports are logged in YouTubeReportHistory including the report createTime which is used to ensure the same report is not collected more than once.
                                       //ReportList All [DataSource] [DownloadPath]
                                       //ReportList [JobID] [DataSource] [DownloadPath]
                                       //ReportList all "Data Source=RG00-SQL01;Integrated Security=true; Initial Catalog=Control; MultipleActiveResultSets=True;" "F:\File\YouTube"

                        if (args[1].ToLower() == "all")
                        { //Get reports for all jobs
                            ListJobsResponse jobListResponseReport = ytReportingService.Jobs.List().Execute();
                            if (jobListResponseReport.Jobs == null)
                            {
                                Console.WriteLine("No Jobs Returned");
                            }
                            else
                            {
                                SqlConnection conn = OpenConnection(args[2]);
                                VerifySchemaYouTubeReportHistory(conn);
                                
                                foreach (Job Job in jobListResponseReport.Jobs)
                                {
                                    Report(Job.Id, conn, args[3], Job.ReportTypeId);
                                }
                            }   
                            break;
                        }
                        else
                        { //Get reports for specified Job
                            SqlConnection conn = OpenConnection(args[2]);
                            VerifySchemaYouTubeReportHistory(conn);
                            Job jobResponse = ytReportingService.Jobs.Get(args[1]).Execute();
                            Report(args[1], conn, args[3], jobResponse.ReportTypeId); 
                        }

                        break;
                    default:
                        break;
                }
            }
        }

        //Merge the collected object into YouTubeOjbect table enabling ID values in reports to be associated with friendly names and descriptions.
        private static void RecordObject(SqlConnection conn, string objectName, string objectID, DateTime publishedAt, string title, string description)
        {
            //Merge Object into dbo.YouTubeObject
            string objectMergeSQL = "Merge into [dbo].[YouTubeObject] tgt" + "\n" +
                "Using" + "\n" +
                "(Select" + "\n" +
                "cast('" + objectName + "' as varchar(250)) [object]," + "\n" +
                "cast('" + objectID + "' as varchar(250)) [objectId]," + "\n" +
                "cast('" + publishedAt.ToString("yyyy - MM - dd HH: mm:ss.fff") + "' as datetime) [publishedAt]," + "\n" +
                "cast('" + title.Replace("'", "''") + "' as varchar(8000)) [title]," + "\n" +
                "cast('" + description.Replace("'", "''") + "' as varchar(8000)) [description]) src on" + "\n" +
                "tgt.[object] = src.[object] and" + "\n" +
                "tgt.[objectId] = src.[objectId]" + "\n" +
                "When not matched then" + "\n" +
                "INSERT" + "\n" +
                "([object]" + "\n" +
                ",[objectId]" + "\n" +
                ",[publishedAt]" + "\n" +
                ",[title]" + "\n" +
                ",[description])" + "\n" +
                "Values (src.[object]" + "\n" +
                ",src.[objectId]" + "\n" +
                ",src.[publishedAt]" + "\n" +
                ",src.[title]" + "\n" +
                ",src.[description])" + "\n" +
                "When matched and" + "\n" +
                "tgt.[publishedAt] <> src.[publishedAt] or (tgt.[publishedAt] is null and  src.[publishedAt] is not null) or (tgt.[publishedAt] is not null and  src.[publishedAt] is null) or" + "\n" +
                "tgt.[title] <> src.[title] or (tgt.[title] is null and src.[title] is not null) or (tgt.[title] is not null and src.[title] is null) or" + "\n" +
                "tgt.[description] <> src.[description] or (tgt.[description] is null and  src.[description] is not null) or (tgt.[description] is not null and  src.[description] is  null)" + "\n" +
                "Then" + "\n" +
                "Update Set" + "\n" +
                "tgt.[loadTime] = sysdatetime()," + "\n" +
                "tgt.[publishedAt] = src.[publishedAt]," + "\n" +
                "tgt.[title] = src.[title]," + "\n" +
                "tgt.[description] = src.[description];";
            SqlCommand objectMerge = new SqlCommand(objectMergeSQL, conn);
            objectMerge.ExecuteNonQuery();
            objectMerge.Dispose();

            //Write to console
            Console.WriteLine("================== Object Recorded ==================");
            Console.WriteLine("Type: " + objectName);
            Console.WriteLine("Id: " + objectID);
            Console.WriteLine("Published At: " + publishedAt.ToString("yyyy-MM-dd HH:mm:ss.fff"));
            Console.WriteLine("Title: " + title);
            Console.WriteLine("Description: " + description);
        }

        //Download the selected report csv, save to specified path, record downlad in YouTubeReportHistory
        private static void Report(string JobId, SqlConnection conn, string filePath, string reportTypeId)
        {
            //Add final backslash if not exists
            if (Right(filePath, 1) != "\\" || Right(filePath, 1) != "/")
            {
                filePath += "\\";
            }
            //Get the last created time to ensure that only reports not yet collected are requested.  If empty then use 1900.
            var reportList = ytReportingService.Jobs.Reports.List(JobId);
            try
            {
                reportList.CreatedAfter = LastCreatedTime(conn, JobId);
            }
            catch
            {
                reportList.CreatedAfter = "1900-01-01T12:00:00.000000Z";
            }
            ListReportsResponse reportListResponse = reportList.Execute();

            if (reportListResponse.Reports == null)
            {
                Console.WriteLine("No Reports for JobID: " + JobId);
            }
            else
            {
                foreach (Report report in reportListResponse.Reports)
                {
                    //Download the report     
                    MediaResource.DownloadRequest getRequest = ytReportingService.Media.Download("");
                    getRequest.BearerToken = getRequest.AccessToken;

                    using (var fileStream = new System.IO.FileStream(@filePath + reportTypeId + "_" + report.Id + ".csv", System.IO.FileMode.Create, System.IO.FileAccess.Write))
                    {
                        // Handler notifies on each chunk download and when the download is completed or failed.
                        getRequest.MediaDownloader.ProgressChanged += Download_ProgressChanged;
                        getRequest.MediaDownloader.Download(report.DownloadUrl, fileStream);
                    }

                    //Insert report into history
                    string reportInsertSql = "INSERT INTO [dbo].[YouTubeReportHistory]" + "\n" +
                   "([reportId]" + "\n" +
                   ",[jobId]" + "\n" +
                   ",[startTime]" + "\n" +
                   ",[endTime]" + "\n" +
                   ",[createTime]" + "\n" +
                   ",[jobExpireTime]" + "\n" +
                   ",[downloadUrl])" + "\n" +
                   "VALUES" + "\n" +
                   "('" + report.Id  + "'\n" +
                   ",'" + JobId + "'\n" +
                   ",'" + ((DateTime)report.StartTime).ToString("yyyy-MM-dd HH:mm:ss.fff") + "'\n" +
                   ",'" + ((DateTime)report.EndTime).ToString("yyyy-MM-dd HH:mm:ss.fff") + "'\n" +
                   ",'" + ((DateTime)report.CreateTime).ToString("yyyy-MM-dd HH:mm:ss.ffffff") + "'\n" +
                   ",'" + report.JobExpireTime + "'\n" + //JobExpreTime is null
                   ",'" + report.DownloadUrl + "')";
                    
                    SqlCommand reportInsert = new SqlCommand(reportInsertSql, conn);
                    reportInsert.ExecuteNonQuery();
                    reportInsert.Dispose();
                    WriteReport(report, "Report");

                    Thread.Sleep(1000); //Free quota is limited to 100 per 100 seconds (throttle to 1 per second max).  
                }
            }
        }

        //Provide user feedback when downloading reports
        static void Download_ProgressChanged(IDownloadProgress progress)
        {
            Console.WriteLine("================== Download Progress ==================");
            Console.WriteLine(progress.Status + " " + progress.BytesDownloaded);
        }

        //Get the last createTime for all previously collected reports for a specified job.
        private static string LastCreatedTime(SqlConnection conn, string JobId)
        {
            //100 nanoseconds is added to createdtime.  When submitting same time as api returned some reports were being collected multiple times.  Assumed to be an issue with precision.
            //Removed 100 nanosecond offset. Download duplicate files is only any issue with downloading the inital set.  Going forward should not be a problem.
            SqlCommand comm = new SqlCommand("SELECT isnull(convert(varchar(50), max(dateadd(NANOSECOND, 0, cast([createTime] as datetimeoffset)) ), 127), '1900-01-01T12:00:00.000000Z') createTimeZulu FROM [dbo].[YouTubeReportHistory] where jobId  = '" + JobId + "'", conn);
            SqlDataReader reader = null;
            reader = comm.ExecuteReader();
            string LastCreatedTime = "";
            if (reader.HasRows)
            {
                reader.Read();
                LastCreatedTime = reader["createTimeZulu"].ToString();
                reader.Close();
                comm.Dispose();
            }
            else
            {
                LastCreatedTime = "1900-01-01T12:00:00.000000Z";
            }
            return LastCreatedTime;
        }

        //Feedback information only
        private static void WriteReport(Report report, String Event)
        {
            Console.WriteLine("================== Job Report ==================");
            Console.WriteLine("Event: " + Event);
            Console.WriteLine("ID: " + report.Id);
            Console.WriteLine("DownloadUrl: " + report.DownloadUrl);
            Console.WriteLine("CreateTime: " + ((DateTime)report.CreateTime).ToString("yyyy-MM-dd HH:mm:ss.fff"));
            Console.WriteLine("StartTime: " + ((DateTime)report.StartTime).ToString("yyyy-MM-dd HH:mm:ss.fff"));
            Console.WriteLine("EndTime: " + ((DateTime)report.EndTime).ToString("yyyy-MM-dd HH:mm:ss.fff"));
            Console.WriteLine("JobID: " + report.JobId);
            Console.WriteLine("JobExpireTime: " + report.JobExpireTime); //ExpireTime is null
        }

        //Feedback information only
        private static void WriteJob (Job job, String Event) {
            Console.WriteLine("================== Reporting Job ==================");
            Console.WriteLine("Event: " + Event);
            Console.WriteLine("ID: " + job.Id);
            Console.WriteLine("Name: " + job.Name);
            Console.WriteLine("ReportTypeID: " + job.ReportTypeId);
            Console.WriteLine("CreateTime: " + ((DateTime)job.CreateTime).ToString("yyyy-MM-dd HH:mm:ss.ffffff"));
            Console.WriteLine("ExpireTime: " + job.ExpireTime); //ExpireTime is null
        }

        //Feedback information only
        private static void WriteReportType(ReportType reportType, String Event)
        {
            Console.WriteLine("================== Report Type ==================");
            Console.WriteLine("Event: " + Event);
            Console.WriteLine("ID: " + reportType.Id);
            Console.WriteLine("Name: " + reportType.Name);
            Console.WriteLine("SystemManaged: " + reportType.SystemManaged);
        }

        //Create a job.  The Data API will generate a csv daily for the created job.
        private static Job CreateJob(String ReportType)
        {
            Job job = new Job();
            job.ReportTypeId = ReportType;
            job.Name = ReportType; //Use the type as the name
            Job createdjob = ytReportingService.Jobs.Create(job).Execute();

            return createdjob;
        }

        private static String DeleteJob(String JobID)
        {
            ytReportingService.Jobs.Delete(JobID);
            return JobID;
        }

        //Establish connection
        private static SqlConnection OpenConnection(string connectionString)
        {
            //Data Source = WIN-IDD5KG7V0RT\SS2016;Integrated Security=true; Initial Catalog=Control; MultipleActiveResultSets=True;
            //Database must exists & MARS = True
            SqlConnection conn = new SqlConnection(connectionString);
            try
            {
                conn.Open();
            }
            catch(Exception e)
            {
                Console.WriteLine(e.ToString());
            }
            return conn;
        }

        //Close connection
        private static void CloseConnection(SqlConnection conn)
        {
            try
            {
                conn.Close();
            }
            catch (Exception e)
            {
                Console.WriteLine(e.ToString());
            }
        }

        //Ensure schema is in place
        private static void VerifySchemaYouTubeReportHistory(SqlConnection conn)
        {
            //Ensure table exists
            //Ensure db exists
            String tblExists = "if not exists (Select * From sys.tables where name = 'YouTubeReportHistory')" + "\n" +
            "Begin" + "\n" +
            "CREATE TABLE[dbo].[YouTubeReportHistory](" + "\n" +
            "[id][int] IDENTITY(1, 1) NOT NULL," + "\n" +
            "[loadTime] [datetime] NOT NULL," + "\n" +
            "[reportId] [varchar](50) NOT NULL," + "\n" +
            "[jobId] [varchar](50) NOT NULL," + "\n" +
            "[startTime] [datetime] NOT NULL," + "\n" +
            "[endTime] [datetime] NOT NULL," + "\n" +
            "[createTime] [datetime2](7) NOT NULL," + "\n" +
            "[jobExpireTime] [datetime] NOT NULL," + "\n" +
            "[downloadUrl] [varchar](1000) NOT NULL," + "\n" +
            "CONSTRAINT[PK_YouTubeReportHistory] PRIMARY KEY CLUSTERED([id] ASC)" + "\n" +
            ")" + "\n" +
            "ALTER TABLE[dbo].[YouTubeReportHistory] ADD CONSTRAINT[DF_YouTubeReportHistory_loadTime]  DEFAULT(sysdatetime()) FOR[loadTime]" + "\n" +
            "End";
            SqlCommand comm = new SqlCommand(tblExists, conn); 
            comm.ExecuteNonQuery();
            comm.Dispose();
        }

        //Ensure schema is in place
        private static void VerifySchemaYouTubeObject(SqlConnection conn)
        {
            //Ensure table exists
            //Ensure db exists
            String tblExists = "if not exists (Select * From sys.tables where name = 'YouTubeObject')" + "\n" +
            "Begin" + "\n" +
            "CREATE TABLE[dbo].[YouTubeObject](" + "\n" +
            "[id][int] IDENTITY(1, 1) NOT NULL," + "\n" +
            "[loadTime] [datetime] NOT NULL," + "\n" +
            "[object] [varchar](250) NOT NULL," + "\n" +
            "[objectId] [varchar](250) NOT NULL," + "\n" +
            "[publishedAt] [datetime] NULL," + "\n" +
            "[title] [varchar](8000) NULL," + "\n" +
            "[description] [varchar](8000) NULL," + "\n" +
            "CONSTRAINT[PK_YouTubeObject] PRIMARY KEY CLUSTERED([id] ASC)" + "\n" +
            ")" + "\n" +
            "ALTER TABLE[dbo].[YouTubeObject] ADD CONSTRAINT[DF_YouTubeObject_loadTime]  DEFAULT(sysdatetime()) FOR[loadTime]" + "\n" +
            "End";
            SqlCommand comm = new SqlCommand(tblExists, conn);
            comm.ExecuteNonQuery();
            comm.Dispose();
        }

        private static UserCredential ytCredential = userCred();
        private static YouTubeReportingService ytReportingService = authYouTubeReportingService();
        private static YouTubeService ytService = authYouTubeService();

        private static UserCredential userCred()
        {
            UserCredential creds;
            using (var stream = new FileStream("client_secret.json", FileMode.Open, FileAccess.Read))
            {
                creds = GoogleWebAuthorizationBroker.AuthorizeAsync(
                    GoogleClientSecrets.Load(stream).Secrets,
                    new[] { YouTubeReportingService.Scope.YtAnalyticsMonetaryReadonly, //Reporting & Analytics API
                            YouTubeReportingService.Scope.YtAnalyticsReadonly, //Reporting & Analytics API
                            YouTubeService.Scope.Youtube, //Data API
                            YouTubeService.Scope.YoutubeForceSsl, //Data API
                            YouTubeService.Scope.YoutubeReadonly, //Data API
                            YouTubeService.Scope.Youtubepartner }, //Data API
                    "user",
                    CancellationToken.None,
                    new FileDataStore("YouTubeAPI") //The refresh token and access token will be stored in application data under a folder with this name
                    ).Result;
            }
            return creds;
        }

        private static YouTubeReportingService authYouTubeReportingService()
        {
            UserCredential creds = ytCredential;
            var service = new YouTubeReportingService(new BaseClientService.Initializer()
            {
                HttpClientInitializer = creds,
                ApplicationName = "LeapFrogBI YouTube Data Collector",
            });
            return service;
        }

        private static YouTubeService authYouTubeService()
        {
            UserCredential creds = ytCredential;
            var ytService = new YouTubeService(new BaseClientService.Initializer()
            {
                HttpClientInitializer = creds,
                ApplicationName = "LeapFrogBI YouTube Data Collector",
            });
            return ytService;
        }

        //Helper - Get substring of specified number of characters on the right.
        private static string Right(string value, int length)
        {
            if (String.IsNullOrEmpty(value)) return string.Empty;
            return value.Length <= length ? value : value.Substring(value.Length - length);
        }
    }


}
